/*****************************************************************************
 * opensles.c : audio output for android native code
 *****************************************************************************
 * Copyright (C) 2010 the VideoLAN team
 *
 * Authors: Dominique Martinet <asmadeus@codewreck.org>
 *          Hugo Beauzée-Luyssen <beauze.h@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_aout.h>
#include <assert.h>

// For native audio
#include <SLES/OpenSLES.h>
#include <SLES/OpenSLES_Android.h>

#define BUFF_QUEUE  10

/*****************************************************************************
 * aout_sys_t: audio output method descriptor
 *****************************************************************************
 * This structure is part of the audio output thread descriptor.
 * It describes the direct sound specific properties of an audio device.
 *****************************************************************************/
struct aout_sys_t
{
    SLObjectItf                     engineObject;
    SLEngineItf                     engineEngine;
    SLObjectItf                     outputMixObject;
    SLAndroidSimpleBufferQueueItf   playerBufferQueue;
    SLObjectItf                     playerObject;
    SLPlayItf                       playerPlay;
    vlc_thread_t                    p_thread;
    vlc_mutex_t                     cond_mutex;
    vlc_cond_t                      cond_wait;
    mtime_t                         first_date;
    vlc_sem_t                       wait_sem;
};

/*****************************************************************************
 * Local prototypes.
 *****************************************************************************/
static int      Open        ( vlc_object_t * );
static void     Close       ( vlc_object_t * );
static void     Play        ( aout_instance_t * );
static void*    PlaybackThreadLoop( aout_instance_t * );
static int      CountAvailableBuffers( SLAndroidSimpleBufferQueueItf pp_buffItf );

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/

vlc_module_begin ()
    set_description( N_("OpenSLES audio output") )
    set_shortname( N_("OpenSLES") )
    set_category( CAT_AUDIO )
    set_subcategory( SUBCAT_AUDIO_AOUT )

    set_capability( "audio output", 170 )
    add_shortcut( "opensles", "android" )
    set_callbacks( Open, Close )
vlc_module_end ()

static void Clear( aout_sys_t *p_sys )
{
    // destroy buffer queue audio player object, and invalidate all associated interfaces
    if ( p_sys->playerObject != NULL )
        (*p_sys->playerObject)->Destroy( p_sys->playerObject );

    // destroy output mix object, and invalidate all associated interfaces
    if ( p_sys->outputMixObject != NULL )
        (*p_sys->outputMixObject)->Destroy( p_sys->outputMixObject );

    // destroy engine object, and invalidate all associated interfaces
    if ( p_sys->engineObject != NULL )
        (*p_sys->engineObject)->Destroy( p_sys->engineObject );

    vlc_cond_destroy( &p_sys->cond_wait );
    vlc_mutex_destroy( &p_sys->cond_mutex );
    vlc_sem_destroy( &p_sys->wait_sem );

    free( p_sys );
}

/*****************************************************************************
 * Open: open a dummy audio device
 *****************************************************************************/
static int Open( vlc_object_t * p_this )
{

    aout_instance_t * p_aout = (aout_instance_t *)p_this;
    SLresult result;

    /* Allocate structure */
    p_aout->output.p_sys = malloc( sizeof( aout_sys_t ) );
    if ( unlikely( p_aout->output.p_sys == NULL ) )
        return VLC_ENOMEM;

    p_aout->output.pf_play = Play;

    aout_sys_t * p_sys = p_aout->output.p_sys;

    p_sys->playerObject=NULL;
    p_sys->engineObject=NULL;
    p_sys->outputMixObject=NULL;
    p_sys->p_thread = NULL;
    vlc_mutex_init( &p_sys->cond_mutex );
    vlc_cond_init( &p_sys->cond_wait );
    p_sys->first_date = 0;
    vlc_sem_init( &p_sys->wait_sem, 0 );

    // create engine
    result = slCreateEngine(&p_sys->engineObject, 0, NULL, 0, NULL, NULL);
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "slCreateEngine error" );
        goto error;
    }

    // realize the engine in synchronous mode
    result = (*p_sys->engineObject)->Realize( p_sys->engineObject, SL_BOOLEAN_FALSE );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "engineObject->Realize error" );
        goto error;
    }

    // get the engine interface, which is needed in order to create other objects
    result = (*p_sys->engineObject)->GetInterface( p_sys->engineObject, SL_IID_ENGINE, &p_sys->engineEngine );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "engineObject->GetInterface error" );
        goto error;
    }

    // create output mix, with environmental reverb specified as a non-required interface
    const SLInterfaceID ids1[] = { SL_IID_VOLUME };
    const SLboolean req1[] = { SL_BOOLEAN_FALSE };
    result = (*p_sys->engineEngine)->CreateOutputMix( p_sys->engineEngine, &p_sys->outputMixObject, 1, ids1, req1 );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err(p_aout, "CreateOutputMix error");
        goto error;
    }

    // realize the output mix in synchronous mode
    result = (*p_sys->outputMixObject)->Realize( p_sys->outputMixObject, SL_BOOLEAN_FALSE );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err(p_aout, "outputMixObject->Realize error");
        goto error;
    }


    // configure audio source - this defines the number of samples you can enqueue.
    SLDataLocator_AndroidSimpleBufferQueue loc_bufq = {
        SL_DATALOCATOR_ANDROIDSIMPLEBUFFERQUEUE,
        BUFF_QUEUE
    };

    SLDataFormat_PCM format_pcm = {
        SL_DATAFORMAT_PCM,
        2,
        SL_SAMPLINGRATE_44_1,
        SL_PCMSAMPLEFORMAT_FIXED_16,
        16,
        SL_SPEAKER_FRONT_RIGHT | SL_SPEAKER_FRONT_LEFT,
        SL_BYTEORDER_LITTLEENDIAN};
    SLDataSource audioSrc = {&loc_bufq, &format_pcm};

    // configure audio sink
    SLDataLocator_OutputMix loc_outmix = { SL_DATALOCATOR_OUTPUTMIX, p_sys->outputMixObject };
    SLDataSink audioSnk = {&loc_outmix, NULL};


    //create audio player
    const SLInterfaceID ids2[] = { SL_IID_ANDROIDSIMPLEBUFFERQUEUE };
    const SLboolean req2[] = { SL_BOOLEAN_TRUE };
    result = (*p_sys->engineEngine)->CreateAudioPlayer( p_sys->engineEngine,
                                                        &p_sys->playerObject, &audioSrc,
                                                        &audioSnk, sizeof( ids2 ) / sizeof( *ids2 ),
                                                        ids2, req2 );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "CreateAudioPlayer error" );
        goto error;
    }

    // realize the player
    result = (*p_sys->playerObject)->Realize( p_sys->playerObject, SL_BOOLEAN_FALSE );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "Realize error" );
        goto error;
    }

    // get the play interface
    result = (*p_sys->playerObject)->GetInterface( p_sys->playerObject, SL_IID_PLAY, &p_sys->playerPlay );
    if (unlikely( result != SL_RESULT_SUCCESS))
    {
        msg_Err( p_aout, "GetInterface SL_IID_PLAY error" );
        return VLC_EGENERIC;
    }

    // get the buffer queue interface
    result = (*p_sys->playerObject)->GetInterface( p_sys->playerObject, SL_IID_ANDROIDSIMPLEBUFFERQUEUE,
                                                    &p_sys->playerBufferQueue );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "GetInterface SL_IID_ANDROIDSIMPLEBUFFERQUEUE error" );
        return VLC_EGENERIC;
    }

    // set the player's state to playing
    result = (*p_sys->playerPlay)->SetPlayState( p_sys->playerPlay, SL_PLAYSTATE_PLAYING );
    if ( unlikely( result != SL_RESULT_SUCCESS ) )
    {
        msg_Err( p_aout, "SetPlayState error" );
        goto error;
    }

    // we want 16bit signed data little endian.
    p_aout->output.output.i_format              = VLC_CODEC_S16L;
    p_aout->output.i_nb_samples                 = 1024;
    p_aout->output.output.i_rate                = 44100;
    p_aout->output.output.i_physical_channels   = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT;

    vlc_value_t val, text;
    val.i_int = AOUT_VAR_STEREO;
    text.psz_string = _("Stereo");
    var_Change( p_aout, "audio-device", VLC_VAR_ADDCHOICE, &val, &text );
    var_Set( p_aout, "audio-device", val );

    aout_FormatPrepare( &p_aout->output.output );

    /* Setting up the playback thread : */
    if ( vlc_clone( &p_sys->p_thread, PlaybackThreadLoop, p_aout,
                           VLC_THREAD_PRIORITY_OUTPUT ) != VLC_SUCCESS )
    {
        msg_Err( p_aout, "Failed to create the output thread" );
        goto error;
    }

    return VLC_SUCCESS;
error:
    Clear( p_sys );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * Close: close our file
 *****************************************************************************/
static void Close( vlc_object_t * p_this )
{
    aout_instance_t *p_aout = (aout_instance_t *)p_this;
    aout_sys_t      *p_sys = p_aout->output.p_sys;

    msg_Dbg( p_aout, "Closing OpenSLES" );

    vlc_object_kill( p_aout );
//    Awake the thread, in case it is still asleep
    vlc_mutex_lock( &p_sys->cond_mutex );
    vlc_cond_signal( &p_sys->cond_wait );
    vlc_mutex_unlock( &p_sys->cond_mutex );

    vlc_cancel( p_sys->p_thread );
    vlc_join( p_sys->p_thread, NULL );

    (*p_sys->playerPlay)->SetPlayState(p_sys->playerPlay, SL_PLAYSTATE_STOPPED );
    //Flush remaining buffers if any.
    if ( p_sys->playerBufferQueue != NULL )
        (*p_sys->playerBufferQueue)->Clear( p_sys->playerBufferQueue );

    Clear( p_sys );
}

/*****************************************************************************
 * Play: play a sound, but actually nothing to do (played in the callback)
 *****************************************************************************/
static void Play( aout_instance_t * p_aout )
{
    aout_sys_t * p_sys = p_aout->output.p_sys;

    vlc_mutex_lock( &p_sys->cond_mutex );
    if ( p_sys->first_date == 0 )
    {
        p_sys->first_date = mdate();
        sem_post( &p_sys->wait_sem );
    }
    vlc_cond_signal( &p_sys->cond_wait );
    vlc_mutex_unlock( &p_sys->cond_mutex );
}

static  int CountAvailableBuffers( SLAndroidSimpleBufferQueueItf pp_buffItf )
{
    SLAndroidSimpleBufferQueueState state;

    SLresult    res = (*pp_buffItf)->GetState( pp_buffItf, &state );
    if ( res != SL_RESULT_SUCCESS )
        return 0;
    return BUFF_QUEUE - state.count;
}

static void* PlaybackThreadLoop( aout_instance_t *p_aout )
{
    aout_sys_t              *p_sys = p_aout->output.p_sys;
    mtime_t                 next_date = mdate();

    vlc_sem_wait( &p_sys->wait_sem );
    mwait( p_sys->first_date - AOUT_PTS_TOLERANCE / 4 );

    while ( vlc_object_alive( p_aout ) )
    {
        vlc_mutex_lock( &p_sys->cond_mutex );
        vlc_cond_wait( &p_sys->cond_wait, &p_sys->cond_mutex );
        vlc_mutex_unlock( &p_sys->cond_mutex );


        if ( !vlc_object_alive( p_aout ) )
            break ;
        int         nbBuffer = CountAvailableBuffers( p_sys->playerBufferQueue );
        mtime_t     currentDate = mdate();
        if ( next_date < currentDate )
            next_date = currentDate;
        while ( nbBuffer-- > 0 )
        {
            aout_buffer_t   *p_buffer = aout_OutputNextBuffer( p_aout, next_date, false );
            if ( p_buffer == NULL )
            {
                if ( p_aout->output.b_starving )
                {
                    //Wait for the half duration of a sample
                    msleep( 500000 / p_aout->output.output.i_rate );
                    break ;
                }
            }

            SLresult    result = (*p_sys->playerBufferQueue)->Enqueue(p_sys->playerBufferQueue, p_buffer->p_buffer, p_buffer->i_buffer);
            if ( result != SL_RESULT_SUCCESS )
            {
                //Don't try to enqueue more buffers...
                aout_BufferFree( p_buffer );
                if (result == SL_RESULT_BUFFER_INSUFFICIENT)
                    msg_Err( p_aout, "write error: buffer insufficient" );
                else
                    msg_Err( p_aout, "write error (%u)", result );
                break ;
            }
            aout_BufferFree( p_buffer );
        }
    }
    return NULL;
}
